use anyhow;
use dirs;
use log::warn;
use serde::{self, Deserialize, Serialize};
use std::{env, path::PathBuf};
use tokio::{fs, io::AsyncReadExt};

use crate::PROGRAM_NAME;

#[derive(Deserialize, Serialize, Debug)]
pub struct Config {
    pub zone_id: Box<str>,
    pub api_key: Box<str>,
    pub domains: Vec<Box<str>>,
    pub max_errors_in_row: Option<usize>,
}

pub async fn get_config_path() -> Result<PathBuf, Vec<PathBuf>> {
    let mut tried_paths = Vec::with_capacity(2);

    match env::current_dir() {
        Ok(current_dir) => {
            let cwd_config = current_dir.join(format!("{}.toml", PROGRAM_NAME));

            let is_valid_path = fs::metadata(&cwd_config)
                .await
                .map_or_else(|_| false, |meta| meta.is_file());

            if is_valid_path {
                return Ok(cwd_config);
            } else {
                tried_paths.push(cwd_config);
            }
        }
        Err(e) => {
            warn!("Failed to get current working directory: {:?}", e);
        }
    }

    match dirs::config_dir() {
        Some(config_dir) => {
            let config_file = config_dir.join(PROGRAM_NAME).join("config.toml");

            let is_valid_path = fs::metadata(&config_file)
                .await
                .map_or_else(|_| false, |meta| meta.is_file());
            if is_valid_path {
                return Ok(config_file);
            } else {
                tried_paths.push(config_file);
            }
        }
        None => {
            warn!("No configuration directory found.");
        }
    }

    Err(tried_paths)
}

pub async fn read_config(path: &PathBuf) -> anyhow::Result<Config> {
    let mut file = fs::File::open(&path).await?;
    let mut buf = String::new();
    file.read_to_string(&mut buf).await?;
    Ok(toml::from_str(&buf)?)
}
